/*  -- LAPACK driver routine (version 3.2) -- */
/*     Univ. of Tennessee, Univ. of California Berkeley and NAG Ltd.. */
/*     November 2006 */

/*  Purpose */
/*  ======= */

/*  DSYEVR computes selected eigenvalues and, optionally, eigenvectors */
/*  of a real symmetric matrix A.  Eigenvalues and eigenvectors can be */
/*  selected by specifying either a range of values or a range of */
/*  indices for the desired eigenvalues. */

/*  DSYEVR first reduces the matrix A to tridiagonal form T with a call */
/*  to DSYTRD.  Then, whenever possible, DSYEVR calls DSTEMR to compute */
/*  the eigenspectrum using Relatively Robust Representations.  DSTEMR */
/*  computes eigenvalues by the dqds algorithm, while orthogonal */
/*  eigenvectors are computed from various "good" L D L^T representations */
/*  (also known as Relatively Robust Representations). Gram-Schmidt */
/*  orthogonalization is avoided as far as possible. More specifically, */
/*  the various steps of the algorithm are as follows. */

/*  For each unreduced block (submatrix) of T, */
/*     (a) Compute T - sigma I  = L D L^T, so that L and D */
/*         define all the wanted eigenvalues to high relative accuracy. */
/*         This means that small relative changes in the entries of D and L */
/*         cause only small relative changes in the eigenvalues and */
/*         eigenvectors. The standard (unfactored) representation of the */
/*         tridiagonal matrix T does not have this property in general. */
/*     (b) Compute the eigenvalues to suitable accuracy. */
/*         If the eigenvectors are desired, the algorithm attains full */
/*         accuracy of the computed eigenvalues only right before */
/*         the corresponding vectors have to be computed, see steps c) and d). */
/*     (c) For each cluster of close eigenvalues, select a new */
/*         shift close to the cluster, find a new factorization, and refine */
/*         the shifted eigenvalues to suitable accuracy. */
/*     (d) For each eigenvalue with a large enough relative separation compute */
/*         the corresponding eigenvector by forming a rank revealing twisted */
/*         factorization. Go back to (c) for any clusters that remain. */

/*  The desired accuracy of the output can be specified by the input */
/*  parameter ABSTOL. */

/*  For more details, see DSTEMR's documentation and: */
/*  - Inderjit S. Dhillon and Beresford N. Parlett: "Multiple representations */
/*    to compute orthogonal eigenvectors of symmetric tridiagonal matrices," */
/*    Linear Algebra and its Applications, 387(1), pp. 1-28, August 2004. */
/*  - Inderjit Dhillon and Beresford Parlett: "Orthogonal Eigenvectors and */
/*    Relative Gaps," SIAM Journal on Matrix Analysis and Applications, Vol. 25, */
/*    2004.  Also LAPACK Working Note 154. */
/*  - Inderjit Dhillon: "A new O(n^2) algorithm for the symmetric */
/*    tridiagonal eigenvalue/eigenvector problem", */
/*    Computer Science Division Technical Report No. UCB/CSD-97-971, */
/*    UC Berkeley, May 1997. */


/*  Note 1 : DSYEVR calls DSTEMR when the full spectrum is requested */
/*  on machines which conform to the ieee-754 floating point standard. */
/*  DSYEVR calls DSTEBZ and SSTEIN on non-ieee machines and */
/*  when partial spectrum requests are made. */

/*  Normal execution of DSTEMR may create NaNs and infinities and */
/*  hence may abort due to a floating point exception in environments */
/*  which do not handle NaNs and infinities in the ieee standard default */
/*  manner. */

/*  Arguments */
/*  ========= */

/*  JOBZ    (input) CHARACTER*1 */
/*          = 'N':  Compute eigenvalues only; */
/*          = 'V':  Compute eigenvalues and eigenvectors. */

/*  RANGE   (input) CHARACTER*1 */
/*          = 'A': all eigenvalues will be found. */
/*          = 'V': all eigenvalues in the half-open interval (VL,VU] */
/*                 will be found. */
/*          = 'I': the IL-th through IU-th eigenvalues will be found. */
/* ********* For RANGE = 'V' or 'I' and IU - IL < N - 1, DSTEBZ and */
/* ********* DSTEIN are called */

/*  UPLO    (input) CHARACTER*1 */
/*          = 'U':  Upper triangle of A is stored; */
/*          = 'L':  Lower triangle of A is stored. */

/*  N       (input) INTEGER */
/*          The order of the matrix A.  N >= 0. */

/*  A       (input/output) DOUBLE PRECISION array, dimension (LDA, N) */
/*          On entry, the symmetric matrix A.  If UPLO = 'U', the */
/*          leading N-by-N upper triangular part of A contains the */
/*          upper triangular part of the matrix A.  If UPLO = 'L', */
/*          the leading N-by-N lower triangular part of A contains */
/*          the lower triangular part of the matrix A. */
/*          On exit, the lower triangle (if UPLO='L') or the upper */
/*          triangle (if UPLO='U') of A, including the diagonal, is */
/*          destroyed. */

/*  LDA     (input) INTEGER */
/*          The leading dimension of the array A.  LDA >= max(1,N). */

/*  VL      (input) DOUBLE PRECISION */
/*  VU      (input) DOUBLE PRECISION */
/*          If RANGE='V', the lower and upper bounds of the interval to */
/*          be searched for eigenvalues. VL < VU. */
/*          Not referenced if RANGE = 'A' or 'I'. */

/*  IL      (input) INTEGER */
/*  IU      (input) INTEGER */
/*          If RANGE='I', the indices (in ascending order) of the */
/*          smallest and largest eigenvalues to be returned. */
/*          1 <= IL <= IU <= N, if N > 0; IL = 1 and IU = 0 if N = 0. */
/*          Not referenced if RANGE = 'A' or 'V'. */

/*  ABSTOL  (input) DOUBLE PRECISION */
/*          The absolute error tolerance for the eigenvalues. */
/*          An approximate eigenvalue is accepted as converged */
/*          when it is determined to lie in an interval [a,b] */
/*          of width less than or equal to */

/*                  ABSTOL + EPS *   max( |a|,|b| ) , */

/*          where EPS is the machine precision.  If ABSTOL is less than */
/*          or equal to zero, then  EPS*|T|  will be used in its place, */
/*          where |T| is the 1-norm of the tridiagonal matrix obtained */
/*          by reducing A to tridiagonal form. */

/*          See "Computing Small Singular Values of Bidiagonal Matrices */
/*          with Guaranteed High Relative Accuracy," by Demmel and */
/*          Kahan, LAPACK Working Note #3. */

/*          If high relative accuracy is important, set ABSTOL to */
/*          DLAMCH( 'Safe minimum' ).  Doing so will guarantee that */
/*          eigenvalues are computed to high relative accuracy when */
/*          possible in future releases.  The current code does not */
/*          make any guarantees about high relative accuracy, but */
/*          future releases will. See J. Barlow and J. Demmel, */
/*          "Computing Accurate Eigensystems of Scaled Diagonally */
/*          Dominant Matrices", LAPACK Working Note #7, for a discussion */
/*          of which matrices define their eigenvalues to high relative */
/*          accuracy. */

/*  M       (output) INTEGER */
/*          The total number of eigenvalues found.  0 <= M <= N. */
/*          If RANGE = 'A', M = N, and if RANGE = 'I', M = IU-IL+1. */

/*  W       (output) DOUBLE PRECISION array, dimension (N) */
/*          The first M elements contain the selected eigenvalues in */
/*          ascending order. */

/*  Z       (output) DOUBLE PRECISION array, dimension (LDZ, max(1,M)) */
/*          If JOBZ = 'V', then if INFO = 0, the first M columns of Z */
/*          contain the orthonormal eigenvectors of the matrix A */
/*          corresponding to the selected eigenvalues, with the i-th */
/*          column of Z holding the eigenvector associated with W(i). */
/*          If JOBZ = 'N', then Z is not referenced. */
/*          Note: the user must ensure that at least max(1,M) columns are */
/*          supplied in the array Z; if RANGE = 'V', the exact value of M */
/*          is not known in advance and an upper bound must be used. */
/*          Supplying N columns is always safe. */

/*  LDZ     (input) INTEGER */
/*          The leading dimension of the array Z.  LDZ >= 1, and if */
/*          JOBZ = 'V', LDZ >= max(1,N). */

/*  ISUPPZ  (output) INTEGER array, dimension ( 2*max(1,M) ) */
/*          The support of the eigenvectors in Z, i.e., the indices */
/*          indicating the nonzero elements in Z. The i-th eigenvector */
/*          is nonzero only in elements ISUPPZ( 2*i-1 ) through */
/*          ISUPPZ( 2*i ). */
/* ********* Implemented only for RANGE = 'A' or 'I' and IU - IL = N - 1 */

/*  WORK    (workspace/output) DOUBLE PRECISION array, dimension (MAX(1,LWORK)) */
/*          On exit, if INFO = 0, WORK(1) returns the optimal LWORK. */

/*  LWORK   (input) INTEGER */
/*          The dimension of the array WORK.  LWORK >= max(1,26*N). */
/*          For optimal efficiency, LWORK >= (NB+6)*N, */
/*          where NB is the max of the blocksize for DSYTRD and DORMTR */
/*          returned by ILAENV. */

/*          If LWORK = -1, then a workspace query is assumed; the routine */
/*          only calculates the optimal size of the WORK array, returns */
/*          this value as the first entry of the WORK array, and no error */
/*          message related to LWORK is issued by XERBLA. */

/*  IWORK   (workspace/output) INTEGER array, dimension (MAX(1,LIWORK)) */
/*          On exit, if INFO = 0, IWORK(1) returns the optimal LWORK. */

/*  LIWORK  (input) INTEGER */
/*          The dimension of the array IWORK.  LIWORK >= max(1,10*N). */

/*          If LIWORK = -1, then a workspace query is assumed; the */
/*          routine only calculates the optimal size of the IWORK array, */
/*          returns this value as the first entry of the IWORK array, and */
/*          no error message related to LIWORK is issued by XERBLA. */

/*  INFO    (output) INTEGER */
/*          = 0:  successful exit */
/*          < 0:  if INFO = -i, the i-th argument had an illegal value */
/*          > 0:  Internal error */

/*  Further Details */
/*  =============== */

/*  Based on contributions by */
/*     Inderjit Dhillon, IBM Almaden, USA */
/*     Osni Marques, LBNL/NERSC, USA */
/*     Ken Stanley, Computer Science Division, University of */
/*       California at Berkeley, USA */
/*     Jason Riedy, Computer Science Division, University of */
/*       California at Berkeley, USA */

/* ===================================================================== */

#include <math.h>
#include <stdbool.h>

#include <blasfeo_d_blas_api.h>



#if defined(FORTRAN_BLAS_API)
#define blasfeo_lapack_dormtr dormtr_
#define blasfeo_lapack_dsytrd dsytrd_
#define blasfeo_lapack_dsyevr dsyevr_
#endif



#define min(x, y) (((x) < (y)) ? (x) : (y))
#define max(x, y) (((x) > (y)) ? (x) : (y))



void dcopy_(int *, double *, int *, double *, int *);
double dlamch_(char *);
double dlansy_(char *, char *, int *, double *, int *, double *);
void blasfeo_lapack_dormtr(char *, char *, char *, int *, int *, double *, int *, double *, double *, int *, double *, int *, int *);
void dscal_(int *, double *, double *, int *);
void dstebz_(char *, char *, int *, double *, double *, int *, int *, double *, double *, double *, int *, int *, double *, int *, int *, double *, int *, int *);
void dstein_(int *, double *, double *, int *, double *, int *, int *, double *, int *, double *, int *, int *, int *);
void dstemr_(char *, char *, int *, double *, double *, double *, double *, int *, int *, int *, double *, double *, int *, int *, int *, bool *, double *, int *, int *, int *, int *);
void dsterf_(int *, double *, double *, int *);
void dswap_(int *, double *, int *, double *, int *);
void blasfeo_lapack_dsytrd(char *, int *, double *, int *, double *, double *, double *, double *, int *, int *);
int ilaenv_(int *, char *, char *, int *, int *, int *, int *);
bool lsame_(char *, char *);
void xerbla_(char *, int *);



void blasfeo_lapack_dsyevr(char *jobz, char *range, char *uplo, int *pn, double *A, int *plda, double *vl, double *vu, int *il, int *iu, double *abstol, int *pm, double *w, double *Z, int *pldz, int *isuppz, double *work, int *lwork, int *iwork, int *liwork, int *info)
	{

	int n = *pn;
	int m = *pm;
	int lda = *plda;
	int ldz = *pldz;

	int i_1 = 1;
	int i_2 = 2;
	int i_3 = 3;
	int i_4 = 4;
	int i_10 = 10;
	int i_m1 = -1;

	bool tryrac;
	int ieeeok, nb, lwmin, liwmin, lwkopt, iscale, indtau, indd, inde, inddd, indee, indwk, llwork, indibl, indisp, indifl, indiwo, iinfo, indwkn, llwrkn, nsplit, imax;
	double safmin, eps, smlnum, bignum, rmin, rmax, abstll, vll, vuu, anrm, sigma, tmp1;
    char order[1];

	int ii, jj, kk;
	int i_t0;
	double d_t0;

    /* Function Body */
    ieeeok = ilaenv_(&i_10, "DSYEVR", "N", &i_1, &i_2, &i_3, &i_4);

    bool lower = lsame_(uplo, "L");
    bool wantz = lsame_(jobz, "V");
    bool alleig = lsame_(range, "A");
    bool valeig = lsame_(range, "V");
    bool indeig = lsame_(range, "I");

    bool lquery = *lwork == -1 || *liwork == -1;

    lwmin = max(1, n*26);
    liwmin = max(1, n*10);

    *info = 0;
    if (! (wantz || lsame_(jobz, "N")))
		{
		*info = -1;
		}
	else if (! (alleig || valeig || indeig))
		{
		*info = -2;
		}
	else if (! (lower || lsame_(uplo, "U")))
		{
		*info = -3;
		}
	else if (n < 0)
		{
		*info = -4;
		}
	else if (lda < max(1,n))
		{
		*info = -6;
		}
	else
		{
		if (valeig)
			{
			if (n > 0 && *vu <= *vl)
				{
				*info = -8;
				}
			}
		else if (indeig)
			{
			if (*il < 1 || *il > max(1,n))
				{
				*info = -9;
				}
			else if (*iu < min(n,*il) || *iu > n)
				{
				*info = -10;
				}
			}
		}
    if (*info == 0)
		{
		if (ldz < 1 || wantz && ldz < n)
			{
			*info = -15;
			}
		else if (*lwork < lwmin && ! lquery)
			{
			*info = -18;
			}
		else if (*liwork < liwmin && ! lquery)
			{
			*info = -20;
			}
		}

    if (*info == 0)
		{
		nb = ilaenv_(&i_1, "DSYTRD", uplo, &n, &i_m1, &i_m1, &i_m1);
		i_t0 = ilaenv_(&i_1, "DORMTR", uplo, &n, &i_m1, &i_m1, &i_m1);
		nb = max(nb, i_t0);
		lwkopt = max((nb+1)*n, lwmin);
		work[0] = (double) lwkopt;
		iwork[0] = liwmin;
		}

    if (*info != 0)
		{
		i_t0 = -(*info);
		xerbla_("DSYEVR", &i_t0);
		return;
		}
	else if (lquery)
		{
		return;
		}

/*     Quick return if possible */

    m = 0;
    if (n == 0)
		{
		work[0] = 1.0;
		return;
		}

    if (n == 1)
		{
		work[0] = 7.0;
		if (alleig || indeig)
			{
			m = 1;
			w[0] = A[0];
			}
		else
			{
			if (*vl < A[0] && *vu >= A[0])
				{
				m = 1;
				w[0] = A[0];
				}
			}
		if (wantz)
			{
			Z[0] = 1.0;
			isuppz[0] = 1;
			isuppz[1] = 1;
			}
		return;
		}

/*     Get machine constants. */

    safmin = dlamch_("Safe minimum");
    eps = dlamch_("Precision");
    smlnum = safmin / eps;
    bignum = 1.0 / smlnum;
    rmin = sqrt(smlnum);
    rmax = min(sqrt(bignum), 1.0 / sqrt(sqrt(safmin)));

/*     Scale matrix to allowable range, if necessary. */

    iscale = 0;
    abstll = *abstol;
	if(valeig)
		{
		vll = *vl;
		vuu = *vu;
		}
    anrm = dlansy_("M", uplo, &n, &A[0], &lda, &work[0]);
    if (anrm > 0.0 && anrm < rmin)
		{
		iscale = 1;
		sigma = rmin / anrm;
		}
	else if (anrm > rmax)
		{
		iscale = 1;
		sigma = rmax / anrm;
		}
    if (iscale == 1)
		{
		if (lower)
			{
			for (jj = 0; jj < n; jj++)
				{
				i_t0 = n - jj;
				dscal_(&i_t0, &sigma, &A[jj + jj * lda], &i_1);
				}
			}
		else
			{
			for (jj = 0; jj < n; jj++)
				{
				i_t0 = jj+1;
				dscal_(&i_t0, &sigma, &A[jj * lda], &i_1);
				}
			}
		if (*abstol > 0.)
			{
			abstll = *abstol * sigma;
			}
		if (valeig)
			{
			vll = *vl * sigma;
			vuu = *vu * sigma;
			}
		}
/*     Initialize indices into workspaces.  Note: The IWORK indices are */
/*     used only if DSTERF or DSTEMR fail. */
/*     WORK(INDTAU:INDTAU+N-1) stores the scalar factors of the */
/*     elementary reflectors used in DSYTRD. */
    indtau = 0;
/*     WORK(INDD:INDD+N-1) stores the tridiagonal's diagonal entries. */
    indd = indtau + n;
/*     WORK(INDE:INDE+N-1) stores the off-diagonal entries of the */
/*     tridiagonal matrix from DSYTRD. */
    inde = indd + n;
/*     WORK(INDDD:INDDD+N-1) is a copy of the diagonal entries over */
/*     -written by DSTEMR (the DSTERF path copies the diagonal to W). */
    inddd = inde + n;
/*     WORK(INDEE:INDEE+N-1) is a copy of the off-diagonal entries over */
/*     -written while computing the eigenvalues in DSTERF and DSTEMR. */
    indee = inddd + n;
/*     INDWK is the starting offset of the left-over workspace, and */
/*     LLWORK is the remaining workspace size. */
    indwk = indee + n;
    llwork = *lwork - indwk;
/*     IWORK(INDIBL:INDIBL+M-1) corresponds to IBLOCK in DSTEBZ and */
/*     stores the block indices of each of the M<=N eigenvalues. */
    indibl = 0;
/*     IWORK(INDISP:INDISP+NSPLIT-1) corresponds to ISPLIT in DSTEBZ and */
/*     stores the starting and finishing indices of each block. */
    indisp = indibl + n;
/*     IWORK(INDIFL:INDIFL+N-1) stores the indices of eigenvectors */
/*     that corresponding to eigenvectors that fail to converge in */
/*     DSTEIN.  This information is discarded; if any fail, the driver */
/*     returns INFO > 0. */
    indifl = indisp + n;
/*     INDIWO is the offset of the remaining integer workspace. */
    indiwo = indifl + n;

/*     Call DSYTRD to reduce symmetric matrix to tridiagonal form. */

    blasfeo_lapack_dsytrd(uplo, &n, &A[0], &lda, &work[indd], &work[inde], &work[indtau], &work[indwk], &llwork, &iinfo);

/*     If all eigenvalues are desired */
/*     then call DSTERF or DSTEMR and DORMTR. */

    if ((alleig || (indeig && *il == 1 && *iu == n) ) && ieeeok == 1)
		{
		if (! wantz)
			{
			dcopy_(&n, &work[indd], &i_1, &w[0], &i_1);
			i_t0 = n - 1;
			dcopy_(&i_t0, &work[inde], &i_1, &work[indee], &i_1);
			dsterf_(&n, &w[0], &work[indee], info);
			}
		else
			{
			i_t0 = n - 1;
			dcopy_(&i_t0, &work[inde], &i_1, &work[indee], &i_1);
			dcopy_(&n, &work[indd], &i_1, &work[inddd], &i_1);

			if (*abstol <= 2.0 * n * eps)
				{
				tryrac = true;
				}
			else
				{
				tryrac = false;
				}
			dstemr_(jobz, "A", &n, &work[inddd], &work[indee], vl, vu, il, iu, &m, &w[0], &Z[0], &ldz, &n, &isuppz[0], &tryrac, &work[indwk], lwork, &iwork[0], liwork, info);



	/*        Apply orthogonal matrix used in reduction to tridiagonal */
	/*        form to eigenvectors returned by DSTEIN. */

			if (wantz && *info == 0)
				{
				indwkn = inde;
				llwrkn = *lwork - indwkn;
				blasfeo_lapack_dormtr("L", uplo, "N", &n, &m, &A[0], &lda, &work[indtau], &Z[0], &ldz, &work[indwkn], &llwrkn, &iinfo);
				}
			}


		if (*info == 0)
			{
	/*           Everything worked.  Skip DSTEBZ/DSTEIN.  IWORK(:) are */
	/*           undefined. */
			m = n;
			goto L30;
			}
			*info = 0;
		}

/*     Otherwise, call DSTEBZ and, if eigenvectors are desired, DSTEIN. */
/*     Also call DSTEBZ and DSTEIN if DSTEMR fails. */

    if (wantz)
		{
		*(unsigned char *)order = 'B';
		}
	else
		{
		*(unsigned char *)order = 'E';
		}
    dstebz_(range, order, &n, &vll, &vuu, il, iu, &abstll, &work[indd], &work[inde], &m, &nsplit, &w[0], &iwork[indibl], &iwork[indisp], &work[indwk], &iwork[indiwo], info);

    if (wantz)
		{
		dstein_(&n, &work[indd], &work[inde], &m, &w[0], &iwork[indibl], &iwork[indisp], &Z[0], &ldz, &work[indwk], &iwork[indiwo], &iwork[indifl], info);

	/*        Apply orthogonal matrix used in reduction to tridiagonal */
	/*        form to eigenvectors returned by DSTEIN. */

		indwkn = inde;
		llwrkn = *lwork - indwkn;
		blasfeo_lapack_dormtr("L", uplo, "N", &n, &m, &A[0], &lda, &work[indtau], &Z[0], &ldz, &work[indwkn], &llwrkn, &iinfo);
		}

/*     If matrix was scaled, then rescale eigenvalues appropriately. */

/*  Jump here if DSTEMR/DSTEIN succeeded. */
L30:
    if (iscale == 1)
		{
		if (*info == 0)
			{
			imax = m;
			}
		else
			{
			imax = *info - 1;
			}
		d_t0 = 1.0 / sigma;
		dscal_(&imax, &d_t0, &w[0], &i_1);
		}

/*     If eigenvalues are not in order, then sort them, along with */
/*     eigenvectors.  Note: We do not sort the IFAIL portion of IWORK. */
/*     It may not be initialized (if DSTEMR/DSTEIN succeeded), and we do */
/*     not return this detailed information to the user. */

    if (wantz)
		{
		for (kk = 0; kk < m-1; kk++)
			{
			ii = 0;
			tmp1 = w[kk];
			for (jj = kk+1; jj < m; jj++)
				{
				if (w[jj] < tmp1)
					{
					ii = jj;
					tmp1 = w[jj];
					}
				}

			if (ii != 0)
				{
				w[ii] = w[kk];
				w[kk] = tmp1;
				dswap_(&n, &Z[ii * ldz], &i_1, &Z[kk * ldz], &i_1);
				}
			}
		}

/*     Set WORK(1) to optimal workspace size. */

    work[0] = (double) lwkopt;
    iwork[0] = liwmin;

    return;

/*     End of DSYEVR */

} /* dsyevr_ */
